from Statistics import AttributeSet


def CalculatePhysicalDamage(damage, attackerStats, defenderStats):
    # type: (AttributeSet, AttributeSet) -> int
    return ((attackerStats.attack + 20) / (defenderStats.defense + 20) * damage)


def CalculateMagicalDamage(damage, attackerStats, defenderStats):
    # type: (AttributeSet, AttributeSet) -> int
    return ((attackerStats.magicAttack + 20) / (defenderStats.magicDefense + 20) * damage)


def CalculateExperienceGain(targetLevel, playerLevel):
    # type: (int, int) -> int
    return 50


def CalculateNextLevelExperience(level):
    # type: (int) -> int
    return level * 250 + 250


def CalculateThreatGain(damage, entityStats):
    # type: (int, AttributeSet) -> int
    modifier = 100.0 + entityStats.t
    modifier /= 100.0
    return int(modifier * damage)


from Objects import Entity, Item, Living


class MapTile(object):

    def __init__(self, x, y, description):
        # type: (int, int, str) -> None
        self.x = x # Where the tile is in world space
        self.y = y # Where the tile is in world space
        self.description = description # Description to send the player when they enter
        self.creatures = [] # List of Living Entities
        self.players = [] # List of Players
        self.items = [] # List of ItemDrops

    def addEntity(self, entity):
        # type: (Entity) -> None
        if isinstance(entity, Living):
            if isinstance(entity, Player):
                self.add_player(entity)
            else:
                self.creatures.append(entity)
        if isinstance(entity, Item):
            self.items.append(entity)

    def removeEntity(self, entity, killed):
        # type: (Entity, bool) -> None
        if isinstance(entity, Living):
            if isinstance(entity, Player):
                self.remove_player(entity)
            else:
                self.remove_creature(entity, killed)
        if isinstance(entity, Item):
            self.items.append(entity)

    def add_player(self, player):
        # type: (Player) -> None
        for player in self.players:
            player.send("{} has entered the area".format(player.name))
        self.players.append(player)
        self.intro_text(player)
        self.modify_player(player)

    def remove_player(self, player):
        # type: (Player) -> None
        self.players.remove(player)
        for player in self.players:
            player.send("{} has left the area".format(player.name))

    def add_creature(self, living):
        # type: (Living) -> None
        self.creatures.append(living)
        for player in self.players:
            player.send("{} has appeared".format(living.name))

    def remove_creature(self, living, killed):
        # type: (Living, bool) -> None
        self.creatures.remove(living)
        for player in self.players:
            if killed:
                player.send("{} was slain".format(living.name))
            else:
                player.send("{} has wandered off".format(living.name))

    def intro_text(self, player):
        # type: (Player) -> None
        player.send(self.description)

    def modify_player(self, player):
        # type: (Player) -> None
        pass # No modifications in a normal tile


class TrapTile(MapTile):

    def __init__(self, x, y, description, damage, trap_name):
        # type: (int, int, str, int) -> None
        self.damage = damage
        self.trap_name = trap_name

        super(MapTile, self).__init__(x, y, description)

    def modify_player(self, player):
        # type: (Player) -> None
        player.damage(self.damage, None)
        player.send("{} hurt you for {} damage".format(self.trap_name, self.damage))


# Example implementation of a trap room that has a specific trap name
class SpikesTrap(TrapTile):
    def __init__(self, x, y, description, damage):
        # type: (int, int, str, int) -> None
        super(TrapTile, self).__init__(x, y, description, damage, "Spikes")


from Common import Packet, PacketResult


# Server Data
class RaceSelectionResult(PacketResult):

    def __init__(self, success, message=None):
        super(RaceSelectionResult, self).__init__(success, message)


class ClassSelectionResult(PacketResult):
    # Packet from server which contains Race Options for the player
    def __init__(self, success, message=None):
        super(ClassSelectionResult, self).__init__(success, message)


class RaceOptions(Packet):
    # Packet from server which contains Race Options for the player
    def __init__(self, races):
        self.races = races


class ClassOptions(Packet):
    # Packet from server which contains Class Option for the player
    def __init__(self, classes):
        self.classes = classes


class CharacterSummaryResponse(Packet):
    # Packet from server which provides a fully fleshed out character to view
    def __init__(self, character):
        self.character = character


class CharacterSummaryBackResponse(Packet):
    def __init__(self, character):
        self.character = character


# Client Requests
class OptionsRequest(Packet):

    # Packet from client to request Race/Class options
    def __init__(self):
        pass


class CharacterSummaryRequest(Packet):

    # Packet from client to request Character Summary View
    def __init__(self):
        pass


class CharacterSummaryBackRequest(Packet):
    def __init__(self):
        pass


class RaceSelection(Packet):

    # Packet from client to set Race on the Server
    def __init__(self, characterRace):
        self.characterRace = characterRace


class ClassSelection(Packet):

    # Packet from client to set Class on the Server
    def __init__(self, characterClass):
        self.characterClass = characterClass



from Statistics import AttributeSet
from threading import Timer
from enum import Enum
from random import randint
from CombatHelpers import CalculateThreatGain, CalculateExperienceGain, CalculatePhysicalDamage, CalculateNextLevelExperience


class Entity(object):

    def __init__(self, name, description, level):
        # type: (str, str, int) -> None
        self.level = level
        self.name = name
        self.description = description
        self.room = None

    def move(self, room):
        # type: (object) -> None
        self.room = room
        room.addEntity(self)


class Item(Entity):

    def __init__(self, name, description, level, value, weight):
        # type: (str, str, int, float, float) -> None
        self.value = value
        self.weight = float
        super(Item, self).__init__(name, description, level)


class Weapon(Item):

    def __init__(self, name, description, level, value, weight, damage, attack_speed, weapon_type, attributes=None):
        # type: (str, str, int, float, float, str, float, int, AttributeSet) -> None
        self.damage = damage # str
        self.attack_speed = attack_speed
        self.weapon_type = weapon_type
        if attributes is None:
            attributes = AttributeSet()
        self.attributes = attributes
        super(Weapon, self).__init__(name, description, level, value, weight)

    def calculate_damage(self):
        pieces = self.damage.split(' ')
        for i in range(0, len(pieces)):
            if 'd' in pieces[i]:
                dice_info = pieces[i].split('d')
                rolls = int(dice_info[0])
                sides = int(dice_info[1])

                total = 0
                for x in range(0, rolls):
                    total += randint(1, sides)

                pieces[i] = total

        formula = ''
        for piece in pieces:
            formula += str(piece)

        print formula
        value = int(eval(formula))
        if value < 1:
            value = 1

        return value


class Shield(Weapon):

    def __init__(self, name, description, level, value, weight, damage, attack_speed, attributes=None):
        # type: (str, str, int, float, float, str, float, AttributeSet) -> None
        super(Shield, self).__init__(name, description, level, value, weight, damage, attack_speed, WeaponType.Shield, attributes)


class WeaponType(Enum):
    Dagger = 1
    Sword = 2
    Axe = 3
    Mace = 4
    Shield = 5
    Staff = 6
    Two_Handed_Sword = 7
    Two_Handed_Axe = 8
    Two_Handed_Mace = 9
    Bow = 10
    Spear = 11


class WeaponSlot(Enum):
    Main_Hand = 1
    Off_Hand = 2


class EquipmentSlot(Enum):
    Main_Hand = 1
    Off_Hand = 2
    Head = 3
    Chest = 4
    Hands = 5
    Legs = 6
    Feet = 7
    Ring1 = 8
    Ring2 = 9
    Necklace = 10
    Bracelets = 11


class ArmorType(Enum):
    Cloth = 1
    Leather = 2
    Chain = 3
    Plate = 4
    Accessory = 5


class Armor(Item):

    def __init__(self, name, description, level, value, weight, armorSlot, armorType, attributes):
        # type: (str, str, int, float, float, EquipmentSlot, ArmorType, AttributeSet) -> None
        self.armorSlot = armorSlot
        self.armorType = armorType
        self.attributes = attributes
        super(Armor, self).__init__(name, description, level, value, weight)


class ItemDrop(object):

    def __init__(self, item, room, owner):
        # type: (Item, object, Entity) -> None
        self.item = item
        self.room = room
        self.owner = owner
        self.pickup_timer_delay = Timer(15.0, self.remove_owner)
        self.pickup_timer_delay.start()

    def pickup(self, entity):
        if self.can_pickup(entity):
            entity.inventory.append(self.item)
            self.pickup_timer_delay.cancel()

    def can_pickup(self, entity):
        if self.owner is None or self.owner is entity:
            return True
        return False

    def remove_owner(self):
        self.owner = None
        for player in self.room.players:
            player.send("Dropped '{}' is no longer protected".format(self.item.name))


class Equipment(object):

    def __init__(self, owner):
        self.owner = owner
        self.main_hand = None # type: Weapon
        self.off_hand = None # type: Weapon
        self.head = None # type: Armor
        self.chest = None # type: Armor
        self.hands = None # type: Armor
        self.legs = None # type: Armor
        self.feet = None # type: Armor
        self.ring1 = None # type: Armor
        self.ring2 = None # type: Armor
        self.necklace = None # type: Armor
        self.bracelets = None # type: Armor

    def checkSlot(self, item, slot):
        # type: (Item, int) -> list
        unequipping = []
        if isinstance(item, Weapon):
            if item.weapon_type > 5:
                unequipping.append("main_hand")
                unequipping.append("off_hand")
            else:
                if (slot == 1):
                    unequipping.append("main_hand")
                elif item.weapon_type == 5 or slot == 2: #Shield or attempting to dual wield
                    unequipping.append("off_hand")
        if isinstance(item, Armor):
            unequipping.append(item.armorSlot.name.lower())
        return unequipping

    def equip(self, item, slot):
        # Returns item(s) previously equipped
        # type: (Item, EquipmentSlot) -> Item
        unequipped = []
        for itemSlot in self.checkSlot(item, slot):
            if item is not None:
                unequipped.append(self.unequip(item))

        setattr(self, slot.name.lower())

        return unequipped

    def unequip(self, slot):
        item = getattr(self, slot, None)
        if item is not None:
            setattr(self, slot, None)
            return item
        return None


class Living(Entity):

    def __init__(self, name, description, level, attributes):
        # type: (str, str, int, AttributeSet) -> None
        self.level = level # type: int
        if attributes is None:
            attributes = AttributeSet()
        self.attributes = attributes # type: AttributeSet
        super(Living, self).__init__(name, description, level)

    def isAlive(self):
        # type: () -> bool
        return self.attributes.health > 0

    def isDead(self):
        return not self.isAlive()

    def attack(self, target):
        # type: (Living) -> None
        if isinstance(target, Living):
            living = target
            # TODO : Calculate damage from equipment or damage moves
            living.damage(CalculatePhysicalDamage(4, self.attributes, target.attributes), self)

    def damage(self, amount, attacker):
        # type: (int, Living) -> None
        self.attributes.health -= amount
        if not self.isAlive():
            # Give damager exp
            pass


class Creature(Living):

    def __init__(self, name, description, level, attributes):
        # type: (str, str, int, AttributeSet) -> None
        self.target = None # type: Living
        self.threat_table = {} # type: dict
        super(Creature, self).__init__(name, description, level, attributes)

    # Do server tick
    def tick(self):
        if self.target is not None:
            pass

    def damage(self, amount, attacker):
        # type: (int, Living) -> None
        self.threat_table[attacker.name] = self.threat_table.get(attacker.name, 0.0) + \
                                           CalculateThreatGain(amount, attacker.attributes)
        super(Creature, self).damage(amount, attacker)


class Monster(Creature):

    def __init__(self, name, description, level, attributes):
        # type: (str, str, int, AttributeSet) -> None
        super(Monster, self).__init__(name, description, level, attributes)

    def tick(self):
        pass



from enum import Enum
from Helpers import firstLetterLowercase


# If you add a value to these enums
# Add the matching camel cased name to Attributes for type intellisense
class Vital(Enum):
    Health = 1
    Mana = 2
    Stamina = 3


class Attribute(Enum):
    Strength = 1
    Constitution = 2
    Dexterity = 3
    Agility = 4
    Wisdom = 5
    Intelligence = 6
    Charisma = 7


class Statistic(Enum):
    Attack = 1
    MagicAttack = 2
    Defense = 3
    MagicDefense = 4
    Dodge = 5
    MagicDodge = 6
    Critical = 7
    Threat = 8


class AttributeSet(object):

    def __init__(self, **kwargs):
        # Vitals
        self.health = 0
        self.maxHealth = 0
        self.mana = 0
        self.maxMana = 0
        self.stamina = 0
        self.maxStamina = 0

        # Attributes
        self.strength = 0
        self.constitution = 0
        self.dexterity = 0
        self.agility = 0
        self.wisdom = 0
        self.intelligence = 0
        self.charisma = 0

        # Stats
        self.attack = 0
        self.magicAttack = 0
        self.defense = 0
        self.magicDefense = 0
        self.dodge = 0
        self.magicDodge = 0
        self.critical = 0
        self.threat = 0

        allowed = []
        for vital in Vital:
            allowed.append(vital.name)
        for attribute in Attribute:
            allowed.append(attribute.name)
        for stat in Statistic:
            allowed.append(stat.name)

        for key, value in kwargs.iteritems():
            if key in allowed:
                self.__setattr__(firstLetterLowercase(key), value)



from LoreleiLib.Packets.Common import Packet
from enum import Enum


class ViewType(Enum):
    Login = 0
    CharacterCreation = 1


class ViewPacket(Packet):

    def __init__(self, viewType):
        # type: (ViewType) -> None
        self.viewType = viewType


from Statistics import AttributeSet
from Objects import Living, Equipment, ArmorType, WeaponType
from CombatHelpers import CalculateNextLevelExperience, CalculateExperienceGain


class CharacterClass(object):

    def __init__(self, name, description, skillList, weaponList, armorList, starterSet=None):
        # type: (str, str, dict, list, list, list) -> None
        self.name = name # type: str
        self.description = description # type: str
        self.skillList = skillList # type: dict<int, list>
        self.weapons = weaponList # type: list<WeaponType>
        self.armor = armorList # type: list<ArmorType>
        self.starterSet = starterSet # type: list<str>


class CharacterRace(object):

    def __init__(self, name, description, attributes):
        # type: (str, str, AttributeSet) -> None
        self.name = name # type: str
        self.description = description # type: str
        self.attributes = attributes # type: AttributeSet


class Character(Living):

    def __init__(self, name, character_race, character_class):
        self.name = name # type: str
        self.character_race = character_race # type: CharacterRace
        self.character_class = character_class # type: CharacterClass
        self.inventory = [] # type: list

        super(Character, self).__init__(name, "A lonely traveller", 999, None)
        self.experience = 0 # type: int
        self.nextLevel = 1000 # type: int
        self.equipment = Equipment(self) # type: Equipment

        # Core + Equipment = Stats used for battle
        self.coreAttributes = AttributeSet(Strength=10, Constitution=10, Dexterity=10, Agility=10, Wisdom=10, Intelligence=10, Charisma=10) # type: AttributeSet
        self.buffs = [] # type: list

    def buildStats(self):
        self.buildCoreAttributes()
        # self.buildEquipmentAttributes()

    def buildCoreAttributes(self):
        for key, value in self.coreAttributes.__dict__.iteritems():
            setattr(self.attributes, key, value)

        if self.character_race is not None:
            for key, value in self.character_race.attributes.__dict__.iteritems():
                setattr(self.attributes, key, getattr(self.attributes, key) + value)

    def buildEquipmentAttributes(self):
        for key, value in self.equipment.__dict__.iteritems():
            setattr(self.attributes, key, getattr(self.attributes, key) + getattr(value.attributes, key))

    def load(self):
        self.loadCoreAttributes()
        self.loadEquipment()
        self.buildStats()

    def loadCoreAttributes(self):
        pass

    def loadEquipment(self):
        pass

    def giveReward(self, target):
        expGain = CalculateExperienceGain(target.level, self.level)
        if (expGain > 0):
            self.experience += expGain
            if self.experience > self.nextLevel:
                self.levelUp()

    def levelUp(self):
        self.level += 1
        self.experience -= self.nextLevel
        self.nextLevel = CalculateNextLevelExperience(self.level)

    def damage(self, amount, attacker):
        if attacker is not None:
            self.send('{} dealt {} damage to you'.format(attacker.name, amount))
            super(Living, self).damage(amount, attacker)

    def send(self, message):
        raise NotImplementedError


from setuptools import setup, find_packages

setup(name='LoreleiLib',
    version='0.1.5',
      description='The text based MUD made by Lorinthio',
      author='Lorinthio',
      author_email='benrandallswg@gmail.com',
      packages=find_packages(),
      install_requires=[
          'twisted',
          'pygame',
          'enum34'
      ],
      zip_safe=False)


from Common import Packet, PacketResult


# Server Data
class LoginResult(PacketResult):

    def __init__(self, success, message=None):
        super(LoginResult, self).__init__(success, message)


class AccountCreateResult(PacketResult):
    def __init__(self, success, message=None):
        super(AccountCreateResult, self).__init__(success, message)


# Client Requests
class LoginAttempt(Packet):

    def __init__(self, username, password):
        self.username = username
        self.password = password


class AccountCreateAttempt(Packet):

    def __init__(self, username, password):
        self.username = username
        self.password = password



